import json
import praw
from .civility_db import CivilityDB
import datetime
from .tools import Tools
import os


class RedditCrawler:
    reddit_auth = json.loads(os.environ.get("REDDIT"))
    reddit = praw.Reddit(user_agent=reddit_auth["user_agent"],
                         client_id=reddit_auth["client_id"], client_secret=reddit_auth["client_secret"],
                         username=reddit_auth["username"], password=reddit_auth["password"])

    @staticmethod
    def get_history(post):
        comment = RedditCrawler.reddit.comment(id=post["reddit_id"])
        trigger, flag = comment.parent(), comment.parent().parent()
        CivilityDB.update(doc=post, collection="bot")
        return trigger, flag

    @staticmethod
    def check_new_score(r_id):
        return RedditCrawler.reddit.comment(id=r_id).score

    @staticmethod
    def find_bot_id(trigger_id):
        bot_posts = RedditPoster.reddit.redditor("gild_civility").comments.new(limit=None)
        for bot_post in bot_posts:
            if bot_post.parent_id.split("_")[1] == trigger_id:
                return bot_post.id
        return "bot_id_not_found"

    class CivilitySearch:
        @staticmethod
        def process_triggers(obj):
            # Determine whether a comment with '/u/Gild-Civility' meets the criteria to create a solicitation post
            rejection = []
            comment_age = datetime.datetime.utcnow() - datetime.datetime.strptime(obj["posted_at"],
                                                                                  "%Y-%m-%d %H:%M:%S")
            tools = CivilityDB.get_tools()
            if comment_age < datetime.timedelta(hours=tools["min_ages"][obj["subreddit"]]):
                rejection.append("age")

            if len("".join(obj["body"].split())) < 50:  # 50 letters minimum (average of ~10 words)
                rejection.append("length")

            if obj["forest_depth"] > 6:
                rejection.append("depth")

            if obj["score"] <= 1:
                rejection.append("score")

            if obj["forest_depth"] == 0:
                rejection.append("no_parent")
            now = Tools.now
            obj.update({"bot_ready": not(len(rejection)),
                        "rejection_reasons": rejection,
                        "created_at": now,
                        "updated_at": now,
                        "bot_id": None
                        })
            print("Civility Identified!!\n" + obj["reddit_id"])
            return obj


# Controller
class RedditPoster(RedditCrawler):
    @staticmethod
    def post_to_reddit(post, reddit_id):
        posts = CivilityDB.get_post_parts()
        trigger = RedditPoster.reddit.comment(id=reddit_id)
        header = posts["header"] + "  \n"
        footer = "  \n" + posts["footer"]
        full_post = header + post + footer
        # print(full_post)  # print for testing... reddit call in the future
        # bot_reddit_id = str(random.randint(1, 1000))  # Unit test

        try:
            trigger.reply(full_post)
        except Exception as error:
            if "deleted" in error.message:
                return "trigger_deleted"
            elif "history now; it's too late" in error.message:
                return "trigger_archived"
            else:
                return {"status_code": 429, "error": error}
                #  trigger.reply(full_post)

        bot_id_error = "bot_id_not_found"
        bot_reddit_id = RedditCrawler.find_bot_id(reddit_id)

        if bot_reddit_id == bot_id_error:
            print(bot_id_error)
        now = Tools.now
        doc = {"created_at": now,
               "updated_at": now,
               "initial_poem": {"header": header,
                                "body": post,
                                "footer": footer},
               "gilding_poem": None,
               "scores": [{"checked_at": now, "score": 1}],
               "reddit_id": bot_reddit_id}

        return CivilityDB.insert(payload=doc, collection="bot"), None



from twilio.rest import Client
import json
from twilio.twiml.messaging_response import MessagingResponse
import os
from .civility_db import CivilityDB


class Twilio:
    twilio_auth = json.loads(os.environ.get("TWILIO"))
    twilio_client = Client(twilio_auth["account_sid"], twilio_auth["auth_token"])

    @staticmethod
    def notification(trigger, flag):
        sep = "\n==========\n"
        return_id = "First, paste the above into your response on its own line:"
        body = return_id + sep + "Flagged for Civlity:\n" + flag.body + sep + "Flagged by:\n" + trigger.body
        for msg in [str(trigger.id), body]:
            message = Twilio.twilio_client.messages.create(
                from_=os.environ.get("TWILIO_FROM"),
                body=msg,
                to=os.environ.get("TWILIO_TO")
            )
        print("Poem Prompter sent:", message.sid)
        return message.sid

    @staticmethod
    def poem_return(response):
        # Get the poem the user sent our Twilio number
        body = response
        trigger_id = body.split('\n')[0]
        reddit_bot_id, reddit_flag_id = CivilityDB.find_comment(trigger_id)

        poem = '  \n'.join(body.split('\n')[1:])

        resp = MessagingResponse()
        resp.message("Thank you for your contribution!")

        return poem, reddit_bot_id, reddit_flag_id



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gildcivility_helpers",
    version="0.1",
    author="Geoffe-Ga",
    author_email="geoffe.gallinger@gmail.com",
    description="Helper modules for the gildcivility reddit bot",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Geoffe-Ga/gildcivility_helpers.git",
    packages=setuptools.find_packages(),
    download_url="https://github.com/Geoffe-Ga/gildcivility_helpers/archive/v_01.tar.gz",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "requests",
        "pymongo",
        "praw",
        "twilio",
        "datetime",
    ],
    python_requires='>=3.7',
)



import datetime
import time


class Tools:
    now = datetime.datetime.utcnow().isoformat(" ")

    @staticmethod
    def strip_parent(list_with_json, json_key):
        # for getting rid of dictionary elements that can't be json'd
        list_without_json = []
        for comment in list_with_json:
            list_without_json.append({k: v for k, v in comment.items() if k != json_key})
        return list_without_json

    @staticmethod
    def wait_for_rate_limit(error):
        for minutes in error.message.split():
            try:
                seconds = int(minutes) * 60
                print("Rate limit exceeded, sleeping for {} seconds".format(seconds))
                time.sleep(seconds)
            except ValueError:
                continue


from pymongo import MongoClient
import os
from .tools import Tools
from bson import ObjectId
import time


class CivilityDB:
    mongo_auth = os.environ.get("MONGO")
    mongo = MongoClient(mongo_auth)
    trigger = mongo.civility.trigger
    flag = mongo.civility.flag
    bot = mongo.civility.bot
    post = mongo.civility.post

    @staticmethod
    def get_post_parts():
        posts = dict()
        for part in CivilityDB.mongo.civility.post.find(CivilityDB.posts_query):
            posts[part["type"]] = part["body"]
            print("Mongo: found", part)
        return posts

    @staticmethod
    def get_tools():
        tools = dict()
        for tool in CivilityDB.mongo.civility.post.find(CivilityDB.tools_query):
            tools[tool["type"]] = tool["body"]
            print("Mongo: found", tool)
        return tools

    posts_query = {"type": {"$in": ["length", "score", "no_parent", "good", "header", "footer"]}, "active": True}
    tools_query = {"type": {"$in": ["subreddits", "min_ages", "bot_reddit_uid"]}, "active": True}
    ready_for_gilder = {"status": "ready", "gilding_poem": None, "notif": {"$ne": None}}

    bot_eval_search_strings = {"ready": {"bot_ready": True, "bot_id": None},
                               "recheck": {"bot_ready": False, "$or": [
                                   {"rejection_reasons": ["age", "score"]},
                                   {"rejection_reasons": "age"}]},
                               "reject": {"$and": [
                                  {"bot_ready": False},
                                  {"bot_id": None},
                                  {"$and": [{"rejection_reasons": {"$ne": ["age", "score"]}},
                                            {"rejection_reasons": {"$ne": "age"}}]}]
                                  }
                               }

    @staticmethod
    def insert(payload, collection):
        # should receive collection as a class (extend enum class in python)
        if len(payload):

            if collection == "trigger":
                print(collection, "Initiating Mongo Dump! (Heh.) Total documents:", len(payload))
                CivilityDB.trigger.insert_many(payload)
            elif collection == "flag":
                print(collection, "Initiating Mongo Dump! (Heh.) Total documents:", len(payload))
                CivilityDB.flag.insert_many(payload)
            elif collection == "bot":
                print(collection, "Initiating Mongo Dump! (Heh.) Total documents: 1")
                bot_id = CivilityDB.bot.insert_one(payload).inserted_id
                CivilityDB.mongo.close()
                return ObjectId(bot_id)
            CivilityDB.mongo.close()

    @staticmethod
    def update(doc, collection, inp=None, new_score=None, poem=None, notification_sid=None, rejection_reasons=None):
        now = Tools.now
        if collection == "trigger":
            doc = CivilityDB.handle_trigger(inp=inp, new_score=new_score, doc=doc, now=now)
            return doc
        if collection == "bot":
            if inp == "notif":
                CivilityDB.handle_notif(doc=doc, notification_sid=notification_sid, now=now, new_score=new_score)
            if inp == "gilding":
                CivilityDB.handle_gilding(doc=doc, now=now, poem=poem)
            return doc

    @staticmethod
    def handle_trigger(doc, inp, new_score, now):
        # Used to determine which document to update
        trigger_id = doc["_id"]
        if inp and not new_score:
            if inp != "trigger_deleted":
                # run this code to add the bot id to the document, showing that it has already had a post made
                doc["bot_id"] = inp
            else:
                # run this code if there was an error during the bot post
                doc["rejection_reasons"].append(inp)
        elif not new_score:
            pass
        elif new_score > 1:
            try:
                doc["rejection_reasons"].remove("score")
            except ValueError:
                print("ERR:", doc["subreddit"], "has the wrong value mapped for min_age")
                pass
            doc["rejection_reasons"].remove("age")
            print("This score improved. " + doc["reddit_id"])
            doc["bot_ready"] = True
        elif new_score <= 1:
            # This function will only be called on documents that have already passed the age test
            doc["rejection_reasons"].remove("age")

        doc["updated_at"] = now

        try:
            CivilityDB.trigger.update_one({"_id": ObjectId(trigger_id)}, {"$set": doc})
        except MongoClient.errors.ServerSelectionTimeoutError:
            print("CivilityDB.errors.ServerSelectionTimeoutError")
            x = 2
            while x < 300:
                print("Write attempt failed for {}. Trying again in {} seconds".format(trigger_id, x))
                time.sleep(x)
                x *= 2
                try:
                    CivilityDB.trigger.update_one({"_id": ObjectId(trigger_id)}, {"$set": doc})
                    return doc
                except MongoClient.errors.ServerSelectionTimeoutError:
                    continue
            print("Database not responding.")

        return doc

    @staticmethod
    def handle_notif(doc, notification_sid, now, new_score):
        CivilityDB.bot.update_one(doc, {"$set":
                                            {"notif": notification_sid,
                                             "updated_at": now},
                                        "$push":
                                            {"scores":
                                                 {"$each":
                                                      [{"checked_at": now, "score": new_score}]
                                                  }
                                             }
                                        }
                                  )

    @staticmethod
    def handle_gilding(doc, poem, now):
        CivilityDB.bot.update_one({"reddit_id": doc},
                                  {"$set": {"status": "done", "gilding_poem": poem, "updated_at": now}})

    @staticmethod
    def find_comment(trigger_id):
        trigger = CivilityDB.trigger.find_one({"reddit_id": trigger_id})
        bot = CivilityDB.bot.find_one({"_id": ObjectId(trigger["bot_id"])})
        return bot["reddit_id"], trigger["flag_id"]
